"""
Contents: Spectral estimators

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Spectral estimators that can be used to create a 2D spectrum from buoy
observations

Classes:

- `None

Functions:

- `mem`, maximum entrophy method
- `mem2`, ...

How To Use This Module
======================
(See the individual functions for details.)

1. Import it: ``import estimators`` or ``from estimators import ...``.
2. call spec2d_from_spec1d to transform a 1D spectrum into a 2D spectrum.
"""

import numpy
from .mem2 import mem2
from .mem import mem
from .loglikelyhood import log_likelyhood
from scipy.ndimage import gaussian_filter
from typing import overload, Dict, List
from roguewave.wavespectra.spectrum2D import WaveSpectrum2D
from roguewave.wavespectra.spectrum1D import WaveSpectrum1D


# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
@overload
def convert_to_2d_spectrum(data: Dict[str, list[WaveSpectrum1D]],
                           number_of_directions: int = 36,
                           method: str = 'mem2', frequency_smoothing=False,
                           smoothing_lengthscale=1) -> Dict[
    str, list[WaveSpectrum2D]]: ...


@overload
def convert_to_2d_spectrum(data: list[WaveSpectrum1D],
                           number_of_directions: int = 36,
                           method: str = 'mem2', frequency_smoothing=False,
                           smoothing_lengthscale=1) -> list[
    WaveSpectrum2D]: ...


@overload
def convert_to_2d_spectrum(data: WaveSpectrum1D,
                           number_of_directions: int = 36,
                           method: str = 'mem2', frequency_smoothing=False,
                           smoothing_lengthscale=1) -> WaveSpectrum2D: ...


@overload
def convert_to_1d_spectrum(data: Dict[str, list[WaveSpectrum2D]]) -> Dict[
    str, list[WaveSpectrum1D]]: ...


@overload
def convert_to_1d_spectrum(data: list[WaveSpectrum2D]) -> list[
    WaveSpectrum1D]: ...


@overload
def convert_to_1d_spectrum(data: WaveSpectrum2D) -> WaveSpectrum1D: ...


# -----------------------------------------------------------------------------
#                              Implementation
# -----------------------------------------------------------------------------
def convert_to_2d_spectrum(data, number_of_directions: int = 36,
                           method: str = 'mem2', frequency_smoothing=False,
                           smoothing_lengthscale=1):
    if isinstance(data, dict):
        out = {}
        for key in data:
            out[key] = convert_to_2d_spectrum(data[key], number_of_directions,
                                              method, frequency_smoothing,
                                              smoothing_lengthscale)
        return out
    elif isinstance(data, list):
        out = []
        for item in data:
            out.append(convert_to_2d_spectrum(item, number_of_directions,
                                              method, frequency_smoothing,
                                              smoothing_lengthscale))
        return out
    elif isinstance(data, WaveSpectrum1D):
        return spec2d_from_spec1d(data,number_of_directions,
                                              method, frequency_smoothing,
                                              smoothing_lengthscale)
    elif isinstance(data, WaveSpectrum2D):
        return data
    else:
        raise Exception('Cannot convert to 2D spectrum')


def convert_to_1d_spectrum(data):
    if isinstance(data, dict):
        out = {}
        for key in data:
            out[key] = convert_to_1d_spectrum(data[key])
        return out
    elif isinstance(data, list):
        out = []
        for item in data:
            out.append(convert_to_1d_spectrum(item))
        return out
    elif isinstance(data, WaveSpectrum1D):
        return data
    elif isinstance(data, WaveSpectrum2D):
        return spec1d_from_spec2d(data)
    else:
        raise Exception('Cannot convert to 2D spectrum')


def spec2d_from_spec1d(spectrum1D: WaveSpectrum1D,
                       number_of_directions: int = 36,
                       method: str = 'mem2', frequency_smoothing=False,
                       smoothing_lengthscale=1) -> WaveSpectrum2D:
    """
    Construct a 2D spectrum based on the 1.5D spectrum and a spectral
    reconstruction method.

    :param number_of_directions: length of the directional vector for the
    2D spectrum. Directions returned are in degrees

    :param method: Choose a method in ['mem','ridge','msem']
        mem: maximum entrophy (in the Boltzmann sense) method
        Lygre, A., & Krogstad, H. E. (1986). Explicit expression and
        fast but tends to create narrow spectra anderroneous secondary peaks.

        mem2: use entrophy (in the Shannon sense) to maximize. Likely
        best method see- Benoit, M. (1993).

        log: solve underdetermined constrained problem using ridge regression
        as regulizer. Performant because efficient solutions to constrained
        quadratic programming problems exist. The spectra compare well to
        nlmen- though are somewhat broader in general. Does not suffer from
        the spurious peak issue of mem. Default method for there reasons.

    REFERENCES:
    Benoit, M. (1993). Practical comparative performance survey of methods
        used for estimating directional wave spectra from heave-pitch-roll data.
        In Coastal Engineering 1992 (pp. 62-75).

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the
        directional distribution in ocean wave spectra.
        Journal of Physical Oceanography, 16(12), 2052-2060.

    """
    direction = numpy.linspace(0, 360, number_of_directions,
                               endpoint=False)

    # Jacobian to transform distribution as function of radian angles into
    # degrees.
    Jacobian = numpy.pi / 180

    a1 = spectrum1D.a1
    b1 = spectrum1D.b1
    a2 = spectrum1D.a2
    b2 = spectrum1D.b2
    e = spectrum1D.e
    if frequency_smoothing:
        e = gaussian_filter(spectrum1D.e, smoothing_lengthscale)
        a1 = gaussian_filter(a1 * spectrum1D.e, smoothing_lengthscale) / e
        a2 = gaussian_filter(a2 * spectrum1D.e, smoothing_lengthscale) / e
        b1 = gaussian_filter(b1 * spectrum1D.e, smoothing_lengthscale) / e
        b2 = gaussian_filter(b2 * spectrum1D.e, smoothing_lengthscale) / e

        scale = spectrum1D.m0() / numpy.trapz(e, spectrum1D.frequency)
        e = e * scale

    if method.lower() in ['maximum_entropy_method', 'mem']:
        # reconstruct the directional distribution using the maximum entropy
        # method.
        directional_distribution = mem(direction * numpy.pi / 180, a1, b1, a2,
                                       b2) * Jacobian
    elif method.lower() in ['log_likelyhood', 'log']:
        directional_distribution = log_likelyhood(direction * numpy.pi / 180,
                                                  a1, b1, a2,
                                                  b2) * Jacobian
    elif method.lower() in ['maximum_entrophy_method2', 'mem2']:
        directional_distribution = mem2(
            direction * numpy.pi / 180, a1, b1, a2,
            b2) * Jacobian
    else:
        raise Exception(f'unsupported spectral estimator method: {method}')

    # We return a 2D wave spectrum object.
    return WaveSpectrum2D(
        frequency=spectrum1D.frequency,
        directions=direction,
        varianceDensity=e[:, None] * directional_distribution,
        timestamp=spectrum1D.timestamp,
        longitude=spectrum1D.longitude,
        latitude=spectrum1D.latitude
    )


def spec1d_from_spec2d(spectrum: WaveSpectrum2D) -> WaveSpectrum1D:
    return WaveSpectrum1D(
        frequency=spectrum.frequency,
        varianceDensity=spectrum.e,
        timestamp=spectrum.timestamp,
        longitude=spectrum.longitude,
        latitude=spectrum.latitude,
        a1=spectrum.a1,
        b1=spectrum.b1,
        a2=spectrum.a2,
        b2=spectrum.b2,
    )



import logging
import sys

# Logging setup
logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
logger.addHandler(logging.NullHandler())

# Caching
import roguewave.filecache as filecache

# External API tools
from .modeldata import sofarspectralapi
from roguewave.modeldata.sofarspectralapi import load_sofar_spectral_file

# model time
from roguewave.modeldata.timebase import TimeSliceLead, \
    timebase_forecast,TimeSliceForecast, TimeSliceEvaluation, TimeSliceAnalysis

# Interpolate
from .interpolate.dataset import interpolate_dataset, tracks_as_dataset
from .interpolate.geometry import TrackSet,Track,Cluster

# Wave spectrum
from .wavespectra.spectrum1D import WaveSpectrum1D
from .wavespectra.spectrum2D import WaveSpectrum2D
from .wavespectra import spectrum1D, spectrum2D

from .wavespectra.estimators import convert_to_1d_spectrum, \
    convert_to_2d_spectrum

# save/load
from .io.io import save, load

# Restart Files:
from roguewave.wavewatch3 import RestartFile
from roguewave.wavewatch3 import open_restart_file, \
    write_partial_restart_file, write_restart_file, \
    reassemble_restart_file_from_parts

# Partitioning
from .wavespectra.partitioning.wavefields import (
    get_bulk_partitions_from_spectral_partitions,
    get_spectral_partitions_from_2dspectra)
from .wavespectra.partitioning.observations import \
    get_bulk_partitions_from_observations, \
    get_spectral_partitions_from_observations

# Model data
from .modeldata.open_remote import open_remote_dataset
from .modeldata.modelinformation import model_timebase, model_valid_time, \
    available_models, list_available_variables
from roguewave.modeldata.extract import extract_from_remote_dataset
from roguewave.colocate.bulk import colocate_model_spotter, \
    colocated_tracks_as_dataset
from roguewave.modeldata.keygeneration import generate_uris



def set_log_to_file(filename, level=logging.INFO):
    logger.addHandler(logging.FileHandler(filename))
    logger.setLevel(level)


def set_level(level):
    if isinstance(level, int):
        logger.setLevel(level)

    elif isinstance(level, str):
        if level == 'debug':
            logger.setLevel(logging.DEBUG)
        elif level == 'info':
            logger.setLevel(logging.INFO)
        elif level == 'warning':
            logger.setLevel(logging.WARNING)
        else:
            raise ValueError(f'unknown logging level {level}')
    else:
        raise ValueError(f'unknown logging level {level}')


def set_log_to_console(level=logging.INFO):
    logger.addHandler(logging.StreamHandler(stream=sys.stdout))
    set_level(level)



"""
Contents: 1D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""

import numpy
from .wavespectrum import WaveSpectrum
from typing import List, Union
from datetime import datetime



class WaveSpectrum1D(WaveSpectrum):
    spectral_density_units = 'm**2/Hertz'

    def __init__(self,
                 frequency: Union[List[float], numpy.ndarray],
                 varianceDensity: Union[List[float], numpy.ndarray],
                 a1: Union[List[float], numpy.ndarray],
                 b1: Union[List[float], numpy.ndarray],
                 a2: Union[List[float], numpy.ndarray],
                 b2: Union[List[float], numpy.ndarray],
                 timestamp: Union[str, datetime, int, float],
                 latitude: Union[float, None],
                 longitude: Union[float, None],
                 **kwargs
                 ):
        super().__init__(frequency=frequency, varianceDensity=varianceDensity,
                         timestamp=timestamp,latitude=latitude,longitude=longitude)
        self._a1 = numpy.array(a1)
        self._b1 = numpy.array(b1)
        self._b2 = numpy.array(b2)
        self._a2 = numpy.array(a2)
        self._e = numpy.array(self.variance_density)

    @WaveSpectrum.variance_density.setter
    def variance_density(self,val:numpy.ndarray):
        self._variance_density = val
        self._e = val

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = (self._range(fmin, fmax)) & numpy.isfinite(self.e)

        return numpy.trapz(
            self.variance_density[range] * self.frequency[range] ** power,
            self.frequency[range])

    def _create_wave_spectrum_input(self) -> dict:
        return {
            "frequency": self.frequency,
            "varianceDensity": self.variance_density,
            "a1": self.a1,
            "b1": self.b1,
            "a2": self.a2,
            "b2": self.b2,
            "timestamp": self.timestamp,
            "latitude": self.latitude,
            "longitude": self.longitude,
        }

    def copy(self) -> "WaveSpectrum1D":
        return WaveSpectrum1D(
            frequency=self.frequency.copy(),
            varianceDensity=self.variance_density.copy(),
            timestamp=self.timestamp,
            latitude=self.latitude,
            longitude=self.longitude,
            a1=self.a1,
            b1=self.b1,
            a2=self.a2,
            b2=self.b2)

    def __add__(self, other:"WaveSpectrum1D")->"WaveSpectrum1D":
        spectrum = self.copy()
        spectrum.variance_density = (spectrum.variance_density +
            other.variance_density)
        return spectrum

    def __sub__(self, other:"WaveSpectrum1D")->"WaveSpectrum1D":
        spectrum = self.copy()
        spectrum.variance_density = (spectrum.variance_density -
            other.variance_density)
        return spectrum

    def __neg__(self, other:"WaveSpectrum1D")->"WaveSpectrum1D":
        spectrum = self.copy()
        spectrum.variance_density = -spectrum.variance_density
        return spectrum


def empty_spectrum1D_like(spectrum: WaveSpectrum1D) -> WaveSpectrum1D:
    return WaveSpectrum1D(
        frequency=spectrum.frequency.copy(),
        varianceDensity=numpy.zeros_like(spectrum.variance_density),
        timestamp=spectrum.timestamp,
        latitude=spectrum.latitude,
        longitude=spectrum.longitude,
        a1=list(spectrum.a1),
        b1=list(spectrum.b1),
        a2=list(spectrum.a2),
        b2=list(spectrum.b2)
    )





from typing import Dict, Tuple, Mapping
import numpy
from pandas import DataFrame
from xarray import Dataset, DataArray
from roguewave.interpolate.dataarray import interpolate_track_data_arrray
from roguewave.interpolate.dataframe import interpolate_dataframe_time
from roguewave.interpolate.geometry import Geometry, convert_to_track_set

def interpolate_dataset(
        data_set: Dataset,
        geometry:Geometry,
        periodic_coordinates: Dict[str,float] = None,
        periodic_data: Dict[str, Tuple[float, float]] = None,
        time_variable_in_dataset: str = 'time',
        longitude_variable_in_dataset: str = 'longitude',
        latitude_variable_in_dataset: str = 'latitude',
    ):
    geometry = convert_to_track_set(
        geometry, data_set[time_variable_in_dataset].values)

    if periodic_coordinates is None:
        periodic_coordinates = {longitude_variable_in_dataset: 360}

    for variable in data_set:
        if 'direction' in str(variable).lower():
            periodic_data = {variable:(360,360)}

    out = {}
    for name, track in geometry.tracks.items():
        points = {
            time_variable_in_dataset: track.time,
            longitude_variable_in_dataset: track.longitude,
            latitude_variable_in_dataset:track.latitude
        }
        out[name] = interpolate_at_points(
            data_set=data_set,
            points=points,
            independent_variable=time_variable_in_dataset,
            periodic_coordinates=periodic_coordinates,
            periodic_data=periodic_data
        ).to_dataframe()
        if out[name].index.tz is None:
            out[name].index = out[name].index.tz_localize('utc')
    return out

def interpolate_at_points(
        data_set: Dataset,
        points: Dict[str, numpy.ndarray],
        independent_variable=None,
        periodic_coordinates: Dict[str, float] = None,
        periodic_data: Dict[str, Tuple[float, float]] = None
) -> Dataset:
    if periodic_data is None:
        periodic_data = {}

    dimensions = data_set.dims
    if independent_variable is None:
        if 'time' in dimensions:
            independent_variable = 'time'
        else:
            independent_variable = dimensions[0]

    return_data_set = Dataset(
        coords={independent_variable: points[independent_variable]}
    )
    for variable in data_set:
        if variable in periodic_data:
            period_data = periodic_data[variable][0]
            discont = periodic_data[variable][1]
        else:
            period_data = None
            discont = None

        return_data_set[variable] = interpolate_track_data_arrray(
            data_set[variable], points, independent_variable,
            periodic_coordinates=periodic_coordinates,
            period_data=period_data,
            discont=discont
        )
    return return_data_set


def tracks_as_dataset(time, drifter_tracks:Mapping[str,DataFrame],subkey=None):
    tracks = {}
    for track_id, track in drifter_tracks.items():
        if subkey is not None:
            _track = track[subkey]
        else:
            _track = track
        variables = list(_track.columns)

        tracks[track_id] = DataArray(
            interpolate_dataframe_time(_track,time),
            coords={'time': time, 'variables':variables},
            dims=['time','variables'],
            name=track_id,
        )

    dataset = Dataset(tracks).to_array(dim='spotter_id')
    return dataset